/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * FWP - Remotely Controlled Robot - 7.Semester
 * 
 * Class Description:
 * 		This class offers the GUI (Graphical User Interface) for the demo
 * 		of the infrared connection between the RCX and the proxy.
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 * @author Willi Tremmel (TremmelW@gmx.de)
 * 
 * System: Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 7T
 */
package proxy.demo;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JTextField;

import proxy.ProxyMessageHandler;
import util.IRMessage;
import util.IRProtocol;
import util.LostConnectionException;


public class ProxyPanel extends JPanel {
	
	/* The message handler */
	protected ProxyMessageHandler messageHandler = null;

	/* The layout of this panel */
	protected GridBagLayout layout = new GridBagLayout();
	
	/* The textfield to enter the distance to move */
	protected JTextField distanceField = new JTextField(10);
	
	/* The combo box to select the direction to drive into */
	protected JComboBox selectDirectionBox = new JComboBox();
	
	/* The textfield to display the last action reply */
	protected JTextField actionField = new JTextField(10);
	
	/* The direction constant strings */
	protected static final String dirForwardString = "Forward";
	protected static final String dirBackwardString = "Backward";
	protected static final String dirRightString = "Right";
	protected static final String dirLeftString = "Left";
	
	/* The message object where we store received IR messages */
	protected IRMessage message = null;
	
	/**
	 * Create a new proxy panel
	 * @param messageHandler, the message handler to use
	 */
	public ProxyPanel(ProxyMessageHandler messageHandler) {
		super();
		
		// Store the message handler
		this.messageHandler = messageHandler;
		
		// Init the combo box
		this.selectDirectionBox.addItem(dirForwardString);
		this.selectDirectionBox.addItem(dirBackwardString);
		this.selectDirectionBox.addItem(dirRightString);
		this.selectDirectionBox.addItem(dirLeftString);
		
		// Initialize the panel
		this.init();
	}
	
	/**
	 * Initialize the panel
	 */
	protected void init() {
		
		// Set the layout for the panel
		this.setLayout(this.layout);
		
		// Create the abstract action we need when we click onto the "Move" Button
		AbstractAction executeMovementAction = new AbstractAction("Move", null) {
		    public void actionPerformed(ActionEvent e) {
		    	try {
		    		doMovement();
		    	} catch (LostConnectionException ex) {
		    		System.err.println("[Error] - " + this.getClass().getName() + " - Lost Connection to RCX");
		    	}
		    }
		};
		
		// Create the abstract action we need when we click onto the "Alive" Button
		AbstractAction executeAliveAction = new AbstractAction("Alive?", null) {
		    public void actionPerformed(ActionEvent e) {
		    	try {
		    		doAlive();
		    	} catch (LostConnectionException ex) {
		    		System.err.println("[Error] - " + this.getClass().getName() + " - Lost Connection to RCX");
		    	}
		    }
		};
		
		// Create the buttons
		JButton moveButton = new JButton(executeMovementAction);
		JButton aliveButton = new JButton(executeAliveAction);
		
		// Arrange everything on the panel's layout
		this.add(this.selectDirectionBox);
		this.addElementToLayout(selectDirectionBox, layout, 0, 0, 1, 1, 0, 20, GridBagConstraints.WEST);
		this.add(this.distanceField);
		this.addElementToLayout(distanceField, layout, 1, 0, 1, 1, 0, 20, GridBagConstraints.WEST);
		this.add(moveButton);
		this.addElementToLayout(moveButton, layout, 2, 0, 1, 1, 0, 20, GridBagConstraints.WEST);
		
		this.add(aliveButton);
		this.addElementToLayout(aliveButton, layout, 1, 1, 1, 1, 0, 20, GridBagConstraints.WEST);
		
		this.add(this.actionField);
		this.addElementToLayout(actionField, layout, 0, 4, 1, 1, 0, 20, GridBagConstraints.WEST);
	}
	
	/**
	 * Called by the abstract action
	 * Sends an "Alive" packet to the RCX
	 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
	 */
	protected void doAlive() throws LostConnectionException {
		this.messageHandler.sendMessageToRCX(IRProtocol.OPCODE_ACTIVE);
		this.receiveMessage();
	}
	
	/**
	 * Called by the abstract action
	 * Execute the selected movement
	 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
	 */
	protected void doMovement() throws LostConnectionException {
		
		try {
			// Grab the distance from the distancefield
			String distanceText = this.distanceField.getText();
			
			// Check if we have a distance
			if(distanceText != null) {
				int distance = Integer.parseInt(distanceText);
				
				// Grab the direction
				String directionText = (String)(this.selectDirectionBox.getSelectedItem());
				
				// Determine which direction we selected and send the proper message to the RCX
				if(directionText.equalsIgnoreCase(dirForwardString)) {
					
					this.messageHandler.sendMessageToRCX(IRProtocol.OPCODE_FORWARD, distance);
					this.receiveMessage();
					
				} else if(directionText.equalsIgnoreCase(dirBackwardString)) {
					
					this.messageHandler.sendMessageToRCX(IRProtocol.OPCODE_BACKWARD, distance);
					this.receiveMessage();
					
				} else if(directionText.equalsIgnoreCase(dirRightString)) {
					
					this.messageHandler.sendMessageToRCX(IRProtocol.OPCODE_RIGHT, distance);
					this.receiveMessage();
					
				} else if(directionText.equalsIgnoreCase(dirLeftString)) {
					
					this.messageHandler.sendMessageToRCX(IRProtocol.OPCODE_LEFT, distance);
					this.receiveMessage();
					
				} else {
					System.err.println("[Error] - " + this.getClass().getName() + " - Unknown direction selected: " + directionText);
				}
				
			} else {
				System.err.println("[Error] - " + this.getClass().getName() + " - No distance entered given !");
			}
		} catch(IllegalArgumentException e) {
			System.err.println("[Error] - " + this.getClass().getName() + " - Input arguments out of range !");
		}
	}	
	
	/**
	 * Wait for a message from the RCX to come in as reply
	 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
	 */
	protected void receiveMessage() throws LostConnectionException {
		
		// Wait for a message from the RCX (blocks till message has been received)
		message = messageHandler.receiveMessageFromRCX();
		
		// Take a look at the opcode in the message and display the proper result
		switch(message.getOpcode()) {
				
			case IRProtocol.OPCODE_ACTIVE:
				actionField.setText("Active");
			break;
				
			case IRProtocol.OPCODE_MOVE_SUCCESS:
				actionField.setText("Successfull: " + new Integer(message.getValue()).toString());
			break;
				
			case IRProtocol.OPCODE_MOVE_ERROR:
				actionField.setText("Error: " + new Integer(message.getValue()).toString());
			break;
				
			default:
				System.err.println("[Error] - " + this.getClass().getName() + " - Received unknown opcode: " + message.getOpcode());
			}
	}
	
    /**
     * Adds the given Component to the given MainLayout with the given GridBagConstraints
     * @param newComp, Component to be added.
     * @param mainLayout, Layout the Component should be added to.
     * @param gridX, X-Value of the Grid
     * @param gridY, Y-Value of the Grid
     * @param gridWidth, Width of the Grid
     * @param gridHeight, Height of the Grid
     * @param weightX
     * @param weightY
     * @param anchor
     */
    protected void addElementToLayout(Component newComp, GridBagLayout mainLayout, int gridX, int gridY, int gridWidth, int gridHeight, int weightX, int weightY, int anchor) {
    	try {
	    	GridBagConstraints mainConstraints = this.buildConstraints(gridX, gridY, gridWidth, gridHeight, weightX, weightY, anchor);
	    	mainLayout.setConstraints(newComp, mainConstraints);
		} catch (Exception e) {
			System.err.println("Adding an element to the panel layout failed!");
		}
    }
    
    /** 
     * Support Method which returns the GridBagLayout Constraint Object
     * based on the input variables. It is used during the creation of the
     * GUI.
     * @param gridX, X-Value of the Grid
     * @param gridY, Y-Value of the Grid
     * @param gridWidth, Width of the Grid
     * @param gridHeight, Height of the Grid
     * @param weightX
     * @param weightY
     * @param anchor
     * */
    protected GridBagConstraints buildConstraints(int gridX, int gridY, int gridWidth, int gridHeight, int weightX, int weightY, int anchor) {
    	GridBagConstraints newConstraints = null;
    	
    	try {
	    	newConstraints = new GridBagConstraints();
	    	newConstraints.gridx = gridX;
	    	newConstraints.gridy = gridY;
	    	newConstraints.gridwidth = gridWidth;
	    	newConstraints.gridheight = gridHeight;
	    	newConstraints.weightx = weightX;
	    	newConstraints.weighty = weightY;
	    	newConstraints.anchor = anchor;
		} catch (Exception e) {
			System.err.println("Creating layout constraints failed!");
		}
		
    	return newConstraints;
    }
}
