/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * FWP - Remotely Controlled Robot - 7.Semester
 * 
 * Class Description:
 * 		This class for the RCX takes care of communication on the RCX side
 * 
 * Attention:
 * 		Due to the limitations of leJOS timeout while writing to the output stream
 * 		is not supported here. Only timeout while reading from the input stream
 * 		is possible but not always needed (i.e. while waiting for a message from the IR Tower
 * 		to know what the next move is)
 * 
 * Error Codes: (code - method - cause)
 * 		EC1 - constructor() - Failure while creating the RCXPort object
 * 		EC2 - sendValueToTower(int, boolean) - Maximum number of retries reached - Lost Connection - Message Lost !
 * 		EC3 - ReceiveMessageHandler.changeState() - RESEND necessary
 * 		EC4 - ReceiveMessageHandler.inspectValue(int) - RESEND necessary - (exists 2 times in this method)
 * 		EC5 - ReceiveMessageHandler.read() - Failure while receiving a message
 * 		EC6 - sendACKToTower() - Maximum number of retries reached - Lost Connection - Message Lost !
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 * @author Willi Tremmel (TremmelW@gmx.de)
 * 
 * System: Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 7T
 */
package rcx;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import josx.platform.rcx.Sound;
import josx.platform.rcx.TextLCD;
import josx.rcxcomm.RCXPort;
import util.IRMessage;
import util.IRProtocol;
import util.LostConnectionException;

public class RCXMessageHandler {

	/* The RCXPort object we use to communicate */
	private RCXPort port = null;
	
	/* The stream for input communication */
	private DataInputStream dInputStream = null;
	
	/* The stream for output communication */
	private DataOutputStream dOutputStream = null;
	
	/* The object taking care of receiving messages from the tower */
	private ReceiveMessageHandler receiveMessageHandler = null;
	
	/* The timeout value */
	private int timeout = -1;
	
	/* The number of retries to send a packet before final error  */
	private int numberOfRetries = -1;
	
	/* The counter where we remember how many retries we already had */
	private int retryCounter = 0;
	private int retryCounterACK = 0;
	
	/* Variable to store received bytes in */
	private int receivedInt = -1;
	private int receivedIntACK = -1;
	
	/* Boolean used to determine if all went well while sending bytes or if we got a RESEND request */
	private boolean transmissionOK = false;
	
	/**
	 * Create the RCXMessageHandler
	 * 	@param timeout, the value till a timeout occurs while reading or writing (in Milliseconds)
	 * @param numberOfRetries, the number of retries after a timeout occured till the program is aborted
	 */
	public RCXMessageHandler(int timeout, int numberOfRetries) {
		
		try {
			// Remember the values
			this.timeout = timeout;
			this.numberOfRetries = numberOfRetries;
			
			// Create the RCX Port
			this.port = new RCXPort();
			
			// Store the input and output streams for this port
			this.dInputStream = new DataInputStream(port.getInputStream());
			this.dOutputStream = new DataOutputStream(port.getOutputStream());
			
			// Create the handler to receive messages
			this.receiveMessageHandler = new ReceiveMessageHandler();
			
		} catch (IOException e) {
			// Display Error - Failure while creating RCXPort Object
			Sound.buzz();
			TextLCD.print("EC1");
			Sound.buzz();
			
			// Unresolvable Error - Exit the Program
			System.exit(0);
		}
	}
	
	/**
	 * Send a certain message to the IR Tower
	 * @param message, the IRMessage containing the opcode and the value to transmit
	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
	 */
	public void sendMessageToTower(IRMessage message) throws LostConnectionException {
		this.sendMessageToTower(message.getOpcode(), message.getValue());
	}
	
	/**
	 * Send a packet with the given opcode and value to IR Tower
	 * @param opcode, the opcode to send
	 * @param value, the value to send with the message
	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
	 */
	public void sendMessageToTower(int opcode, int value) throws LostConnectionException {
		
		// Reset the boolean
		this.transmissionOK = false;
		
		// Do again in case we have a transmission error
		while(!this.transmissionOK) {
			
			// Send the opcode + value
			this.transmissionOK = this.sendValueToTower(opcode + value);
		}
	}
	
	/**
	 * Send a packet with a sole opcode to the IR tower
	 * @param opcode, the opcode to send
	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
	 */
	public void sendMessageToTower(int opcode) throws LostConnectionException {

		// Call the proper method
		this.sendMessageToTower(opcode, 0);
	}
	
	/**
	 * Causes the handler to wait for a message from the tower
	 * Blocks till a full message (opcode and value(if necessary)) has been received
	 * @return the received message
	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
	 */
	public IRMessage receiveMessageFromTower() throws LostConnectionException {
		return this.receiveMessageHandler.read();
	}
	
	/**
	 * Used to send an opcode directly to the tower
	 * @param value, the value to transmit to the tower
	 * @return boolean, true if all went well, false if we got a resend request
	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
	 */
	private boolean sendValueToTower(int value) throws LostConnectionException {
			
		// Send the opcode till the tower responds with an ACK
		this.receivedInt = -1;
		this.retryCounter = 0;
		
		// Keep running till either we received another reply than ACK or the number of retries after timeouts is depleted
		while(this.receivedInt != IRProtocol.OPCODE_MESSAGE_ACK && retryCounter <= this.numberOfRetries) {
			
			try {
				// Send the opcode
				this.dOutputStream.writeInt(value);
				this.dOutputStream.flush();
				
				// Wait for an ACK reply till the timeout is up (this timeout also throws an IOException when happening)
				this.port.setTimeOut(this.timeout);
				this.receivedInt = this.dInputStream.readInt();
				this.port.setTimeOut(0);
				
			} catch (IOException ex) {
				
				// A timeout occured, raise the retry counter and resend
				this.retryCounter++;
			}
		}
		
		// Check if we stopped due to reaching the maximum number of retries
		if(this.retryCounter > this.numberOfRetries) {
			
			// Display Error - Maximum number of retries reached
			Sound.buzz();
			TextLCD.print("EC2");
			
			// Throw the exception so the super class get's notified of that error
			throw new LostConnectionException();
		}
		
		return true;
	}	
	
	/**
	 * Used to send an ACK directly to the tower
	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
	 */
	private boolean sendACKToTower() throws LostConnectionException {
			
		// Send the opcode till the tower responds with an ACK
		this.receivedIntACK = -1;
		this.retryCounterACK = 0;
		
		// Keep running till either we received another reply than ACK or the number of retries after timeouts is depleted
		while(this.receivedIntACK != IRProtocol.OPCODE_MESSAGE_ACK && retryCounterACK <= this.numberOfRetries) {
			
			try {
				// Send the opcode
				this.dOutputStream.writeInt(IRProtocol.OPCODE_MESSAGE_ACK);
				this.dOutputStream.flush();
				
				// We won't get a reply to an ACK opcode we sent, so make sure we leave the while(...)
				this.receivedIntACK = IRProtocol.OPCODE_MESSAGE_ACK;
					
			} catch (IOException ex) {
				
				// A timeout occured, raise the retry counter and resend
				this.retryCounterACK++;
			}
		}
		
		// Check if we stopped due to reaching the maximum number of retries
		if(this.retryCounterACK > this.numberOfRetries) {
			
			// Display Error - Maximum number of retries reached
			Sound.buzz();
			TextLCD.print("EC6");
			
			// Throw the exception so the super class get's notified of that error
			throw new LostConnectionException();
		}
		
		return true;
	}	
	
	
	/**
	 * Subclass taking care of receiving messages
	 */
	class ReceiveMessageHandler {
				
		/* Variables to store data */
		private int readValue = -1;
		
		/* Tells us if we already have the full message or not */
		private boolean messageIncomplete = true;
		
		/* The message object we reply with when reading */
		private IRMessage message = new IRMessage();
		
		/**
		 * Construct a new ReceiveMessageHandler
		 */
		public ReceiveMessageHandler() {}
		
		/**
		 * Read a message coming in from the IR Tower
		 * This method blocks till a full message (opcode and if needed a value) has been received
		 * @return the received message object
		 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
		 */
		public IRMessage read() throws LostConnectionException {
				
			try {
				// Reset the boolean
				this.messageIncomplete = true;
				
				// Reset all other values
				this.reset();
				
				// Keep reading till the boolean indicates that the message is complete
				while(this.messageIncomplete) {
					
					// Read a value from the input stream (no timeout)
					port.setTimeOut(0);
					this.readValue = dInputStream.readInt();
					
					// Acknowledge this message in case it was no ACK we received
					if(readValue != IRProtocol.OPCODE_MESSAGE_ACK) {
						
						// No ACK, send one
						boolean sendACKSuccess = sendACKToTower();
						
						if(!sendACKSuccess) {
							Sound.buzz();
							TextLCD.print("ACK");
						}
					}
					
					// Divide opcode and value
					int distance = readValue%1000;
					int opcode = readValue - distance;
					
					// Build the IRMessage
					this.message = new IRMessage(opcode, distance);
					
					this.messageIncomplete = false;
				}
				
				// Make the audio that we got a complete message
				Sound.twoBeeps();
				
				// For debugging purposes, print the received message to the LCD display
				//this.printMessageToLCD();

				// Return the complete message to the calling method
				return this.message;
				
			} catch (IOException e) {
				
				// Display Error - Failure while receiving a message
				Sound.buzz();
				TextLCD.print("EC5");
				
				return null;
			}
		}
		
		/**
		 * Reset everything
		 * Used when we know that we will receive the whole last message again
		 */
		private void reset() {
			this.message.setOpcode(-1);
			this.message.setValue(-1);
		}
		
		
		/**
		 * Print the contents of the current message to the LCD display
		 * For debugging purposes
		 */
		private void printMessageToLCD() {
			
			// Print the Opcode to the LCD
			TextLCD.print(Integer.toString(this.message.getOpcode()));
			
			try {
				// Wait for two seconds
				Thread.sleep(2000);
			} catch (InterruptedException e) {}
			
			// Print the Value to the LCD
			TextLCD.print(Integer.toString(this.message.getValue()));
		}
	}
}

	