/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * FWP - Remotely Controlled Robot - 7.Semester
 * 
 * Class Description:
 * 		This class takes care of actually moving the RCX
 * 
 * Error Codes: (code - method - cause)
 * 		EB1 - backward(int) - Thread Interruption
 * 		EB2 - forward(int) - Thread Interruption
 * 		EB3 - left(int) - Thread Interruption
 * 		EB4 - right(int) - Thread Interruption
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 * @author Willi Tremmel (TremmelW@gmx.de)
 * 
 * System: Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 7T
 */
package rcx;

import josx.platform.rcx.Motor;
import josx.platform.rcx.Sensor;
import josx.platform.rcx.SensorListener;
import josx.platform.rcx.Sound;
import josx.platform.rcx.TextLCD;
import util.IRMessage;
import util.IRProtocol;

public class RCXMovementHandler implements SensorListener {

	/* Tells our main thread if the sensor state changed */
	protected boolean sensorStateChanged = false;
	
	/* The message we respond with, which should be relayed to the IR tower */
	protected IRMessage message = new IRMessage();
	
	/* Last movement opcode  */
	protected int lastMoveOpcode = IRProtocol.OPCODE_FORWARD;
	
	/* Constants to transform the real distance into the ticks we need */
	protected static final double FB_DISTANCE_TRANSFORMATION = 1.618105522054778;
	protected static final double RL_DISTANCE_TRANSFORMATION = 1.5592884;
	
	/* Time we need to put the primary wheels down */
	protected static final int pWheelsDownTime = 1000;
	
	/* Time we need to put the secondary wheels down and the primary ones up */
	protected static final int sWheelsDownTime = 1500;
	
	/* The distance to move away from an obstacle if we hit one (in cm) */
	protected int moveAwayDistance = 2;
	
	/* Local variable to store specific value, allocated here to save memory */
	protected int oldRotValue = -1;
	protected int targetRotValue = -1;
	protected int reachedDistance = -1;
	
	/**
	 * Construct a new DefaultMovementHandler
	 */
	public RCXMovementHandler() {
		
		// Add us as SensorListener to the touch sensors so we get notified when we bump into something
		Sensor.S2.addSensorListener(this);
	}
	
	/**
	 * SensorListener Interface Implementation
	 * This method is called when a sensor, which we added ourself as listener to, changes it's value
	 * @param sensor, the sensor which changed it's value
	 * @param oldValue, the previous sensor value
	 * @param newValue, the new sensor value
	 */
	public void stateChanged(Sensor sensor, int oldValue, int newValue) {
		
		// If the new value of the touch sensor is "1" we know that we bumped into something
		if(newValue == 1) {
			
			// Store the information that a sensor value changed
			this.sensorStateChanged = true;
		}
	}
	
	/**
	 * This method takes care of driving backward for the given distance
	 * @param distance, the distance to drive backward (in centimeters)
	 * @return IRMessage, the return message we need to relay back to the server to indicate successfull movement or an error
	 */
	public IRMessage backward(int distance) {
		
		// Reset all variables
		this.sensorStateChanged = false;
		this.message.setOpcode(-1);
		this.message.setValue(-1);

		// Check if the correct set of wheels is currently on the floor
		if(this.lastMoveOpcode != IRProtocol.OPCODE_FORWARD && this.lastMoveOpcode != IRProtocol.OPCODE_BACKWARD) {
			
			// Wrong wheels on the floor, put them down
			Motor.B.forward();
			
			try {
				Thread.sleep(pWheelsDownTime);
			} catch (InterruptedException e) {
				// Display Error - Thread Interrupted
				Sound.buzz();
				TextLCD.print("EB1");
			}
			
			Motor.B.stop();
		}
		
		// Remember the current Sensor1 value
		this.oldRotValue = Sensor.S1.readValue();
		
		// Calculate the target rotational sensor value we want to reach based on the distance
		this.targetRotValue = Sensor.S1.readValue() - this.transformDistanceToTicks(IRProtocol.OPCODE_BACKWARD, distance);
		
		// Reset the boolean again which indicates a sensor value change
		this.sensorStateChanged = false;
		
		// Initiate the engine to drive backward
		Motor.A.backward();
		
		// Drive the desired distance
		while(targetRotValue < Sensor.S1.readValue() && !this.sensorStateChanged) {
			// Do nothing
		}
		
		// Stop the engine
		Motor.A.stop();
		
		// Calculate the traveled distance
		this.reachedDistance = this.transformDistanceToCentimeters(IRProtocol.OPCODE_BACKWARD, oldRotValue - Sensor.S1.readValue());
		
		// Check if we stopped due to a change of the touch sensor status
		if(this.sensorStateChanged) {
		    
		    // Move a little bit away from the obstacle
		    this.oldRotValue = Sensor.S1.readValue();
		    this.targetRotValue = Sensor.S1.readValue() + this.transformDistanceToTicks(IRProtocol.OPCODE_FORWARD, this.moveAwayDistance);
		    
		    Motor.A.forward();
		    
		    while(targetRotValue > Sensor.S1.readValue()) {
		        // Do nothing
		    }
		    
		    Motor.A.stop();
		    this.reachedDistance = this.reachedDistance - this.moveAwayDistance;
			
			// Create the error IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_ERROR);
			this.message.setValue(this.reachedDistance);
			
		} else {
			
			// Create the success IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
			this.message.setValue(this.reachedDistance);
		}
		
		// Remember this opcode for the next turn		
		this.lastMoveOpcode = IRProtocol.OPCODE_BACKWARD;
		
		/**
		 * Comment this part in for testing purposes
		 * Simply returns a success packet with the given distance
		 * You need to comment the above part out for that purpose except
		 * the lines under "// Reset all variables"
		 */
//		// TEST: Create a dummy message for testing purposes
//		this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
//		this.message.setValue(distance);
//		
//		// Do a beep
//		Sound.beep();
		
		// Return the actual IR Message
    	return this.message;
	}
	
	/**
	 * This method takes care of driving forward for the given distance
	 * @param distance, the distance to drive forward (in centimeters)
	 * @return IRMessage, the return message we need to relay back to the server to indicate successfull movement or an error
	 */
	public IRMessage forward(int distance) {		

		// Reset all variables
		this.sensorStateChanged = false;
		this.message.setOpcode(-1);
		this.message.setValue(-1);
		
		// Check if the correct set of wheels is currently on the floor
		if(this.lastMoveOpcode != IRProtocol.OPCODE_FORWARD && this.lastMoveOpcode != IRProtocol.OPCODE_BACKWARD) {
			
			// Wrong wheels on the floor, put them down
			Motor.B.forward();
			
			try {
				Thread.sleep(pWheelsDownTime);
			} catch (InterruptedException e) {
				// Display Error - Thread Interrupted
				Sound.buzz();
				TextLCD.print("EB2");
			}
			
			Motor.B.stop();
		}
		
		// Remember the current Sensor1 value
		this.oldRotValue = Sensor.S1.readValue();
		
		// Calculate the target rotational sensor value we want to reach based on the distance
		this.targetRotValue = Sensor.S1.readValue() + this.transformDistanceToTicks(IRProtocol.OPCODE_FORWARD, distance);
		
		// Reset the boolean again which indicates a sensor value change
		this.sensorStateChanged = false;
		
		// Initiate the engine to drive backward
		Motor.A.forward();
		
		// Drive the desired distance
		while(targetRotValue > Sensor.S1.readValue() && !this.sensorStateChanged) {
			// Do nothing
		}
		
		// Stop the engine
		Motor.A.stop();
		
		// Calculate the traveled distance
		this.reachedDistance = this.transformDistanceToCentimeters(IRProtocol.OPCODE_FORWARD, Sensor.S1.readValue() - oldRotValue);
		
		// Check if we stopped due to a change of the touch sensor status
		if(this.sensorStateChanged) {
		    
		    // Move a little bit away from the obstacle
		    this.oldRotValue = Sensor.S1.readValue();
		    this.targetRotValue = Sensor.S1.readValue() - this.transformDistanceToTicks(IRProtocol.OPCODE_FORWARD, this.moveAwayDistance);
		    
		    Motor.A.backward();
		    
		    while(targetRotValue < Sensor.S1.readValue()) {
		        // Do nothing
		    }
		    
		    Motor.A.stop();
		    this.reachedDistance = this.reachedDistance - this.moveAwayDistance;
			
			// Create the error IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_ERROR);
			this.message.setValue(this.reachedDistance);
			
		} else {
			
			// Create the success IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
			this.message.setValue(this.reachedDistance);
		}
		
		// Remember this opcode for the next turn		
		this.lastMoveOpcode = IRProtocol.OPCODE_FORWARD;
		
		/**
		 * Comment this part in for testing purposes
		 * Simply returns a success packet with the given distance
		 * You need to comment the above part out for that purpose except
		 * the lines under "// Reset all variables"
		 */
//		// TEST: Create a dummy message for testing purposes
//		this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
//		this.message.setValue(distance);
//		
//		// Do a beep
//		Sound.beep();
		
		// Return the actual IR Message
    	return this.message;
	}
	
	/**
	 * This method takes care of driving left for the given distance
	 * @param distance, the distance to drive left (in centimeters)
	 * @return IRMessage, the return message we need to relay back to the server to indicate successfull movement or an error
	 */
	public IRMessage left(int distance) {		

		// Reset all variables
		this.sensorStateChanged = false;
		this.message.setOpcode(-1);
		this.message.setValue(-1);
		
		// Check if the correct set of wheels is currently on the floor
		if(this.lastMoveOpcode != IRProtocol.OPCODE_LEFT && this.lastMoveOpcode != IRProtocol.OPCODE_RIGHT) {
			
			// Wrong wheels on the floor, put them down
			Motor.B.backward();
			
			try {
				Thread.sleep(sWheelsDownTime);
			} catch (InterruptedException e) {
				// Display Error - Thread Interrupted
				Sound.buzz();
				TextLCD.print("EB3");
			}
			
			Motor.B.stop();
		}
		
		// Remember the current Sensor3 value
		this.oldRotValue = Sensor.S3.readValue();
		
		// Calculate the target rotational sensor value we want to reach based on the distance
		this.targetRotValue = Sensor.S3.readValue() + this.transformDistanceToTicks(IRProtocol.OPCODE_LEFT, distance);
		
		// Reset the boolean again which indicates a sensor value change
		this.sensorStateChanged = false;
		
		// Initiate the engine to drive backward
		Motor.C.forward();
		
		// Drive the desired distance
		while(targetRotValue > Sensor.S3.readValue() && !this.sensorStateChanged) {
			// Do nothing
		}
		
		// Stop the engine
		Motor.C.stop();
		
		// Calculate the traveled distance
		this.reachedDistance = this.transformDistanceToCentimeters(IRProtocol.OPCODE_LEFT, Sensor.S3.readValue() - oldRotValue);
		
		// Check if we stopped due to a change of the touch sensor status
		if(this.sensorStateChanged) {
		    
		    // Move a little bit away from the obstacle
		    this.oldRotValue = Sensor.S3.readValue();
		    this.targetRotValue = Sensor.S3.readValue() - this.transformDistanceToTicks(IRProtocol.OPCODE_FORWARD, this.moveAwayDistance);
		    
		    Motor.C.backward();
		    
		    while(targetRotValue < Sensor.S3.readValue()) {
		        // Do nothing
		    }
		    
		    Motor.C.stop();
		    this.reachedDistance = this.reachedDistance - this.moveAwayDistance;
			
			// Create the error IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_ERROR);
			this.message.setValue(this.reachedDistance);
			
		} else {
			
			// Create the success IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
			this.message.setValue(this.reachedDistance);
		}
		
		// Remember this opcode for the next turn		
		this.lastMoveOpcode = IRProtocol.OPCODE_LEFT;
		
		/**
		 * Comment this part in for testing purposes
		 * Simply returns a success packet with the given distance
		 * You need to comment the above part out for that purpose except
		 * the lines under "// Reset all variables"
		 */
//		// TEST: Create a dummy message for testing purposes
//		this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
//		this.message.setValue(distance);
//		
//		// Do a beep
//		Sound.beep();
		
		// Return the actual IR Message
    	return this.message;
	}
	
	/**
	 * This method takes care of driving right for the given distance
	 * @param distance, the distance to drive right (in centimeters)
	 * @return IRMessage, the return message we need to relay back to the server to indicate successfull movement or an error
	 */
	public IRMessage right(int distance) {

		// Reset all variables
		this.sensorStateChanged = false;
		this.message.setOpcode(-1);
		this.message.setValue(-1);
		
		// Check if the correct set of wheels is currently on the floor
		if(this.lastMoveOpcode != IRProtocol.OPCODE_LEFT && this.lastMoveOpcode != IRProtocol.OPCODE_RIGHT) {
			
			// Wrong wheels on the floor, put them down
			Motor.B.backward();
			
			try {
				Thread.sleep(sWheelsDownTime);
			} catch (InterruptedException e) {
				// Display Error - Thread Interrupted
				Sound.buzz();
				TextLCD.print("EB4");
			}
			
			Motor.B.stop();
		}
		
		// Remember the current Sensor3 value
		this.oldRotValue = Sensor.S3.readValue();
		
		// Calculate the target rotational sensor value we want to reach based on the distance
		this.targetRotValue = Sensor.S3.readValue() - this.transformDistanceToTicks(IRProtocol.OPCODE_RIGHT, distance);
		
		// Reset the boolean again which indicates a sensor value change
		this.sensorStateChanged = false;
		
		// Initiate the engine to drive backward
		Motor.C.backward();
		
		// Drive the desired distance
		while(targetRotValue < Sensor.S3.readValue() && !this.sensorStateChanged) {
			// Do nothing
		}
		
		// Stop the engine
		Motor.C.stop();
		
		// Calculate the traveled distance
		this.reachedDistance = this.transformDistanceToCentimeters(IRProtocol.OPCODE_RIGHT, oldRotValue - Sensor.S3.readValue());
		
		// Check if we stopped due to a change of the touch sensor status
		if(this.sensorStateChanged) {
		    
		    // Move a little bit away from the obstacle
		    this.oldRotValue = Sensor.S3.readValue();
		    this.targetRotValue = Sensor.S3.readValue() + this.transformDistanceToTicks(IRProtocol.OPCODE_FORWARD, this.moveAwayDistance);
		    
		    Motor.C.forward();
		    
		    while(targetRotValue > Sensor.S3.readValue()) {
		        // Do nothing
		    }
		    
		    Motor.C.stop();
		    this.reachedDistance = this.reachedDistance - this.moveAwayDistance;
			
			// Create the error IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_ERROR);
			this.message.setValue(this.reachedDistance);
			
		} else {
			
			// Create the success IR Message
			this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
			this.message.setValue(this.reachedDistance);
		}
		
		// Remember this opcode for the next turn		
		this.lastMoveOpcode = IRProtocol.OPCODE_RIGHT;
		
		/**
		 * Comment this part in for testing purposes
		 * Simply returns a success packet with the given distance
		 * You need to comment the above part out for that purpose except
		 * the lines under "// Reset all variables"
		 */
//		// TEST: Create a dummy message for testing purposes
//		this.message.setOpcode(IRProtocol.OPCODE_MOVE_SUCCESS);
//		this.message.setValue(distance);
//		
//		// Do a beep
//		Sound.beep();
		
		// Return the actual IR Message
    	return this.message;
	}
	
	/**
	 * Transform the distance from centimeters into rotational sensor ticks
	 * @param moveOpcode, to identify the direction we are moving into
	 * @param centimeters, the centimeters to move
	 * @return the ticks, the number of ticks which are needed to drive the given centimeters
	 */
	private int transformDistanceToTicks(int moveOpcode, int centimeters) {
		
		// Are we driving forward/backward or left/right?
		if(moveOpcode == IRProtocol.OPCODE_FORWARD || moveOpcode == IRProtocol.OPCODE_BACKWARD) {
				
			// We are driving forward/backward, use the proper constant for the transformation
			return this.round(centimeters * FB_DISTANCE_TRANSFORMATION);
				
		} else {
			
			// We are driving right/left, use the proper constant for the transformation
			return this.round(centimeters * RL_DISTANCE_TRANSFORMATION);
		}
	}
	
	/**
	 * Transform the distance from rotational sensor ticks into centimeters
	 * @param moveOpcode, to identify the direction we are moving into
	 * @param ticks, the ticks we moved
	 * @return the centimeters, the distance in centimeters we moved
	 */
	private int transformDistanceToCentimeters(int moveOpcode, int ticks) {
		
		// Are we driving forward/backward or left/right?
		if(moveOpcode == IRProtocol.OPCODE_FORWARD || moveOpcode == IRProtocol.OPCODE_BACKWARD) {
					
				// We are driving forward/backward, use the proper constant for the transformation
				return this.round(ticks / FB_DISTANCE_TRANSFORMATION);
					
			} else {
				
				// We are driving right/left, use the proper constant for the transformation
				return this.round(ticks / RL_DISTANCE_TRANSFORMATION);
			}
	}
	
	/**
	 * Self written round() Method. We need this one cause leJOS does not support
	 * the original Math.round() method.
	 * @param value, the value to round to the nearest integer
	 * @return numbers smaller than #.5 will be rounded down, all others (#.5 till #.9) will be rounded up
	 */
	private int round(double value) {
		
		// Replacement for the "%" operator, cause this one is not supported in leJOS for double values
		double moduloValue = value;
		
		// Calculate value%1.0
		while(moduloValue > 1.0) {
			moduloValue = moduloValue - 1.0;
		}
		
		// Check if what is left after doing value%1.0 is smaller than 0.5 or not
		if(moduloValue < 0.5) {
			
			// The left value is smaller than 0.5 so round down
			return (int)value;
		} else {
			
			// The left value is equal or greater than 0.5 so round up
			return (int)(value) +1;
		}
	}
}
